/*
	EZ-FPGA Hnadling Library Ver.1.00

	Copyright(C) 2003 T&A Tstsuyuki Satoh

	1. License

	This program is freeware for non-commercial use. 
	The user of this source must assume all legal 
	responsibility in the use of this program. 

	2. Guarantee 

	An author does no guarantee about this program.
	You can use this under your own risk. 

	3. A rule to distribute.
		Do either 1) or 2).
		1)Include a perfect original source.
		2)Include all these headers , and add document of 
		modification points

	Histry
		2003. 8. 5 bugfix: jmem address not update when size=0
		2003. 7.29 Created
*/

#include <windows.h>
#include <stdio.h>
#include <conio.h>

#include "ezfpga.h"

#if !EZFPGA_STATIC_LIB

/* DLL Driver FilePath */
#define DLL_NAME "lib\\ezfpga"

// DLL Export prefix
#define DLLEXPORT __declspec(dllexport)
#ifdef __MINGW32__
#define DLLAPI
#else
#define DLLAPI WINAPI
#endif

/* EZFPGA DLL Interface */
static DWORD (DLLAPI *ezfpga_init_)(DWORD port_type);
DWORD (DLLAPI *ezfpga_config)(const char *cfg_data,DWORD *remain);
DWORD (DLLAPI *ezfpga_spi_xfer)(BYTE port,BYTE *txbuf,BYTE *rxbuf,DWORD *num_bytes);

/* option */
DWORD (DLLAPI *ezfpga_shutdown)(void);
DWORD (DLLAPI *ezfpga_get_status)(DWORD *deviceid,DWORD *status,DWORD *usercode);
DWORD (DLLAPI *ezfpga_free)(void);

/***************************************************************************/
/* DLL Handling                                                            */
/***************************************************************************/

static HINSTANCE dllHandle;

// get DLL function
static int get_dll_func(void *func_ptr,char *name)
{
	*(void **)func_ptr = (void *)GetProcAddress(dllHandle,name);

	if( *(void **)func_ptr == NULL)
	{
		printf("DLL function '%s' was not found\n",name);
		return 0;
	}
	return 1;
}

// Link DLL
DWORD ezfpga_init(DWORD port_type)
{
	/* Load DLL */
	dllHandle = LoadLibraryEx(DLL_NAME,NULL,0);
	if(dllHandle == NULL)
	{
		printf("Can't open '%s.DLL'\n",DLL_NAME);
		return ERR_DRIVER;
	}

	if(
		get_dll_func(&ezfpga_shutdown,   "ezfpga_shutdown"   ) &
		get_dll_func(&ezfpga_config,     "ezfpga_config"     ) &
		get_dll_func(&ezfpga_spi_xfer,   "ezfpga_spi_xfer"   ) &
		get_dll_func(&ezfpga_init_,       "ezfpga_init"       ) &
// option
		get_dll_func(&ezfpga_get_status, "ezfpga_get_status" ) &
		get_dll_func(&ezfpga_free,       "ezfpga_free"       )
// option
	)
	{
		return ezfpga_init_(port_type);
	}

	/* NG */
	FreeLibrary(dllHandle);
	dllHandle = NULL;
	return ERR_DRIVER;
}

void ezfpga_unlink(void)
{
	ezfpga_spi_xfer(0,NULL,NULL,NULL);
	ezfpga_free();
	if( dllHandle != NULL)
		FreeLibrary(dllHandle);
}
#else

/* static lick */
void ezfpga_unlink(void)
{
	ezfpga_spi_xfer(0,NULL,NULL,NULL);
	ezfpga_free();
}
#endif /* EZFPGA_STATIC_LIB */

#if EZFPGA_CONSOLE_FUNC
/***************************************************************************/
/* Configration with show remain                                           */
/***************************************************************************/
DWORD ezfpga_config_show(const char *config_data,DWORD size)
{
	DWORD err;
	/* set Thread Flag */
	size |= EZFPGA_NON_BLOCK;
	err = ezfpga_config(config_data,&size);
	/* wait and show satus */
	while(1)
	{
		printf("\rConfigration remain=%8ld",size);
		if(!(err&ERR_BUSY))
			break;
		Sleep(100);
		err = ezfpga_config(NULL,NULL); // check again
	}
	printf("\n");
	return err;
}

DWORD ezfpga_xfer_show(BYTE port,BYTE *txbuf,BYTE *rxbuf,DWORD size)
{
	DWORD err;
	/* set Thread Flag */
	size |= EZFPGA_NON_BLOCK;
	err = ezfpga_spi_xfer(1,txbuf,rxbuf,&size);

	while(1)
	{
		printf("\rXFER remain=%8ld",size);
		if(!(err&ERR_BUSY))
			break;
		Sleep(100);
		err = ezfpga_config(NULL,NULL); // check again
	}
	printf("\n");
	return err;
}
#endif /* EZFPGA_CONSOLE_FUNC */

#if EZFPGA_JMEM_FUNC
/***************************************************************************/
/* SRAM Accessing                                                          */
/***************************************************************************/

#define BLOCK_SIZE (0x10000)
#define BLOCK_SIZE_WR (BLOCK_SIZE/0x40)*0x3f

#define JMEM_WR 0x00
#define JMEM_RD 0x01
#define JMEM_VF 0x03

static BYTE tx_buf[BLOCK_SIZE] , rx_buf[BLOCK_SIZE];

DWORD jmem_access(BYTE mode,DWORD address,BYTE *data,DWORD size)
{
	DWORD asize , bsize , bcnt , ssize , errpos;
	DWORD err;

	while(1)
	{
		// address increment limit
		asize = 0x10000-(address&0xffff);
		if(asize>size)
			asize = size;

		// set address
		tx_buf[0x00] = 0x80 | (BYTE)((address>>24)&0x7f);
		tx_buf[0x01] = (BYTE)(address>>16);
		tx_buf[0x02] = (BYTE)(address>>8);
		tx_buf[0x03] = (BYTE)address;
		if(mode==JMEM_WR)
		{	/* write */
			ssize = 4;
		}
		else
		{	/* read / verify */
			tx_buf[0x04] = 0x21; // read inc.
			tx_buf[0x05] = 0x21; // read inc.
			ssize = 6;
		}
		// tranfser address & prefetch read
		err = ezfpga_spi_xfer(1,tx_buf,NULL,&ssize);
		if(err)
			return err;

		// only addres set ?
		if(size==0)
			return 0;


		if(mode!=JMEM_WR)
		{
			// setup read increment commands
			memset(tx_buf,0x21,BLOCK_SIZE);
		}

		bcnt=0;

		while(asize)
		{
			if(mode==JMEM_WR)
			{
				for(ssize=0;asize && ssize<(BLOCK_SIZE-0x40);ssize += bsize + 1)
				{
					bsize = asize>0x3f ? 0x3f : asize;
					tx_buf[ssize] = 0x40 + (BYTE)bsize; // burst write command
					memcpy(&tx_buf[ssize+1],data,bsize);
					size  -= bsize;
					asize -= bsize;
					bcnt  += bsize;
					data  += bsize;
				}
				ezfpga_spi_xfer(1,tx_buf,NULL,&ssize);
			}
			else
			{
				bsize = asize>BLOCK_SIZE ? BLOCK_SIZE : asize;
				ssize = bsize;
				if(mode==JMEM_RD)
				{
					ezfpga_spi_xfer(1,tx_buf,data,&ssize);
				}
				if(mode==JMEM_VF)
				{
					ezfpga_spi_xfer(1,tx_buf,rx_buf,&ssize);
					if(memcmp(data,rx_buf,bsize))
					{
						for(errpos=0;errpos<bsize;errpos++)
						{
							if(data[errpos] != rx_buf[errpos])
							{
								printf("Verify error %08lX\n",address+errpos);
								break;
							}
						}
						return ERR_CRC;
					}
				}
				//
				size  -= bsize;
				asize -= bsize;
				bcnt  += bsize;
				data  += bsize;
			}
			//
			while(bcnt>8192)
			{
				bcnt -= 8192;
				printf(".");
			}
		}
		if(mode==JMEM_WR)
		{
			// Last write cycle
			tx_buf[0x00] = 0x00; // NOP
			ssize = 1;
			ezfpga_spi_xfer(1,tx_buf,NULL,&ssize);
		}
		// next addrss block
		address = (address + 0x10000) & 0xffff0000;
	}
	return 0;
}

DWORD jmem_write(DWORD address,const BYTE *data,DWORD size)
{
	return jmem_access(JMEM_WR,address,(BYTE *)data,size);
}

DWORD jmem_read(DWORD address,BYTE *data,DWORD size)
{
	return jmem_access(JMEM_RD,address,data,size);
}

DWORD jmem_verify(DWORD address,BYTE *data,DWORD size)
{
	return jmem_access(JMEM_VF,address,data,size);
}
#endif /* EZFPGA_JMEM_FUNC */
