/*
	EZ-FPGA + PBSRAM PROM emulator

	Copyright(C) 2003 Tatsutuki Satoh

	2003. 8.20 Ver.1.00 Support Trigger Function
	2003. 7.22 Create (based from chamrome.c)

	note:

	Use GCC-Mingw32

*/
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>

#include "ezfpga.h"
#include "romemuf.c" /* config data */

#define RETRY_CNT 10

/* ROMt@Cpobt@ */
static BYTE data_buf[0x80000];

/* setup register bits */
#define A13EN 0x01
#define A14EN 0x02
#define A15EN 0x04
#define A16EN 0x08
#define A17EN 0x10
#define A18EN 0x20 /* A18 == pin31 */
#define F18EN 0x40 /* A18 == pin1  */

/* ROM types */
#define SEL_2764  0x00
#define SEL_27128 (A13EN)
#define SEL_27256 (A13EN|A14EN)
#define SEL_27512 (A13EN|A14EN|A15EN)
#define SEL_27010 (A13EN|A14EN|A15EN|A16EN)
#define SEL_27020 (A13EN|A14EN|A15EN|A16EN|A17EN)
#define SEL_27040 (A13EN|A14EN|A15EN|A16EN|A17EN|A18EN)
#define SEL_29040 (A13EN|A14EN|A15EN|A16EN|A17EN|F18EN)

#define MAX_ROM_TYPES 8

static const BYTE rom_value[MAX_ROM_TYPES] =
{	SEL_2764 , SEL_27128 , SEL_27256 , SEL_27512 , SEL_27010 , SEL_27020 , SEL_27040 , SEL_29040};

static const char *rom_name[MAX_ROM_TYPES] =
{	"2764" , "27128" , "27256","27512","27010","27020","27040","29040" };

static const DWORD rom_size[MAX_ROM_TYPES] =
{	0x2000 , 0x4000 , 0x8000 , 0x10000 , 0x20000 , 0x40000 , 0x80000 , 0x80000 };

int str_to_romtype(char *string)
{
	int i;

	for(i=0;i<MAX_ROM_TYPES;i++)
	{
		if(strcmp(string,rom_name[i])==0)
		{
			printf("ROM type %s[%d]\n",rom_name[i],i);
			return i;
		}
	}
	return -1;
}

DWORD file_read(char *file_name,void *buf , DWORD size)
{
	HANDLE hndl;
	DWORD read_size;

	// file open
	hndl = CreateFile(file_name,GENERIC_READ,
		0,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0);
	if( hndl==INVALID_HANDLE_VALUE)
		return 0;

	if( !ReadFile(hndl,buf,size,&read_size,NULL) )
	{
		CloseHandle(hndl);
		return 0;
	}
	CloseHandle(hndl);
	return read_size;
}

int prome_run_mode(int run,int rom_type,DWORD trigger_address,int reset_level)
{
	BYTE cntrl_val;
	BYTE adr_data[3];

	/* controll register value */
	cntrl_val = run ? 0x80 : 0x00;

	/* reset controll */
	switch(reset_level|0x20)
	{
	case 'l': cntrl_val |= 0x40; break;
	case 'h': cntrl_val |= 0x60; break;
/*	case 'z': cntrl_val |= 0x00; break; */
	}

	/* ROM type select */
	jmem_write(0x40000007,&rom_value[rom_type],1);

	/* Trigger set */
	if(trigger_address & 0x80000000)
	{
		/* single or pulse */
		if(trigger_address & 0x40000000)
			cntrl_val |= 0x10;

		/* set trigger address */
		adr_data[0] = (BYTE)trigger_address;
		adr_data[1] = (BYTE)(trigger_address>>8);
		adr_data[2] = (BYTE)(trigger_address>>16);
		jmem_write(0x40000001 , adr_data , 3);
	}

	/* set controll register */
	jmem_write(0x40000006 , &cntrl_val , 1);

	return 0;
}

int prome_clear_trigger(void)
{
	char  dmy;

	/* clear trigger */
	jmem_write(0x40000004 , &dmy , 1); /* dummy port */

	return 0;
}

int prome_init(int clock_mhz)
{
	unsigned char buf[16];
	int timeout;

	if( ezfpga_init(DEVICE_USB2(0)))
	{
		printf("Can't found EZ-FPGA\n");
		return 1;
	}

#if 1
	/* CPLD check! */
	ezfpga_spi_xfer(0,NULL,NULL,NULL); /* init jmem */
	jmem_read(0x40000000,buf,16);
	ezfpga_spi_xfer(0,NULL,NULL,NULL); /* init jmem */

	if(buf[12]=='R' && buf[13]=='E' && buf[14]=='F')
	{
		printf("FPGA Already Running\n");
		printf("Trigger= %02X%02X%02X\n",buf[1],buf[2],buf[3]);
		printf("F      = %dMHz\n",buf[5]+1);
		printf("ST     = %02X\n",buf[6]);
		printf("TYPE   = %02X\n",buf[7]);
	}
	else
#endif
	{
		/* Configration */
		if(ezfpga_config_show(romemuf_bit,sizeof(romemuf_bit)))
		{
			printf("Can't Config EZ-FPGA\n");
			ezfpga_unlink();
			return 1;
		}
	}

	/* FPGA interface check */
	jmem_read(0x4000000c,buf,4);
#if 0
	printf("%02X %02X %02X %02X \n",
		buf[0],
		buf[1],
		buf[2],
		buf[3]);
#endif
	if(buf[0]!='R' || buf[1]!='E'|| buf[2]!='F')
	{
		printf("BAD FPGA !\n");
		ezfpga_unlink();
		return 1;
	}

	/* setup PLL */
	printf("set clock = %dMHz:",clock_mhz);
	buf[0] = clock_mhz-1;
	jmem_write(0x40000005,buf,1);

	/* wait for lock PLL & DLL */
	for(timeout=500;timeout;timeout--)
	{
		jmem_read(0x40000004,buf,1);
		if((buf[0]&0x0c)==0x0c)
			break;
		printf(".");
		sleep(10);
	}
	if(timeout==0)
	{
		printf("PLL & DLL unlock error\n");
		ezfpga_unlink();
		return 1;
	}
	printf("Locked!\n");

	return 0;
}

void prome_end(void)
{
	ezfpga_unlink();
}

/* -------------------- finish -------------------- */
int escape(int ret_code,int wait_time)
{
	if(wait_time>0)
	{
		printf("Waiting %dmsec.\n",wait_time);
		sleep(wait_time);
	} else if(wait_time==0)
	{
		printf("HIT ANY KEY\n");
		getch();
	}
	exit(ret_code);
}

/* -------------------- MAIN -------------------- */


// C
int main(int argc,char *argv[])
{
	DWORD file_size;
	char filename[128];

	int i,err;
	int retry;
	int hexa_mode , rom_type , all_mode;
	BYTE pad_data;
	BYTE ver_mode;
	DWORD trigger_mode;
	unsigned int trg_a;
	int reset_on,reset_off;
	int wait_time;
	int clk_mhz;

	// ^Cg\
	printf("\n\
EZ-FPGA PROM emulator Version 1.00\n\
Copyrght(C) 2003 T&A Tatsuuki Satoh\n");

	if(argc<2)
	{
		//@g
		printf("\
Usage: ezfrome [filename] [options]\n\
  filename : Binary emuration ROM image\n\
  options  :\n\
    -r[type]   : select ROM type\n\
                   2764,27128,27256,27512,27010,27020,27040 or 29040\n\
    -a         : write all range of ROM size\n\
    -p[pad]    : set PAD code\n\
    -v         : Verify only\n\
    -ts[addr]  : Single Trigger\n\
    -tp[addr]  : Pulse Trigger\n\
    -tc        : Clear Single Trigger\n\
    -s[on][off]: Reset Output Control\n\
               : -sLZ = OpenDrain Active LOW(Default)\n\
               : -sLH = Active LOW\n\
               : -sHL = Active HIGH\n\
    -c[clkMHz] : PBSRAM Clock Frequency\n\
    -w[msec]   : Wait After Finish,0=KEY Wait\n\
");
		escape(1,0);
	}

	// R}hC̉
	hexa_mode = 0;
	all_mode = 0;
 	rom_type = -1;
	pad_data = 0xff;
	ver_mode = 0;
	filename[0] = 0x00;
	trigger_mode    = 0;
	reset_on  = 'l';
	reset_off = 'z';
	wait_time = -1;
	clk_mhz = 100;

	for( i=1 ; i<argc ; i++ )
	{
		if( argv[i][0] == '/' || argv[i][0] == '-')
		{
			// IvVXCb`
			switch( argv[i][1]|0x20 )
			{
			case 'a': all_mode  = 1; break;
			case 'p': pad_data  = atoi(&argv[i][2]); break;
			case 'h': hexa_mode = 1; break;
			case 'r': rom_type  = str_to_romtype(&argv[i][2]); break;
			case 'v': ver_mode  = 1; break;
			case 't': 
				switch(argv[i][2]|0x20)
				{
				case 's': trigger_mode = 0x80000000; break;
				case 'p': trigger_mode = 0xc0000000; break;
				case 'c': trigger_mode = 0x20000000; break; // clear
				}
				if(trigger_mode&0x80000000)
				{
					if(sscanf(&argv[i][3],"%x",&trg_a))
						trigger_mode |= trg_a;
					else
						trigger_mode = 0;
				}
				break;
			case 's':
				reset_on  = argv[i][2]|0x20;
				if(argv[i][2])
					reset_off = argv[i][3]|0x20;
				break;
			case 'w':
				wait_time = 0;
				sscanf(&argv[i][2],"%d",&wait_time);
				break;
			case 'c': 
				sscanf(&argv[i][2],"%d",&clk_mhz);
				if(clk_mhz>133)
					clk_mhz = 133;
				if(clk_mhz<50)
					clk_mhz = 50;
				break;
			}
		}
		else
		{
			// t@C
			sprintf(filename,"%s",argv[i]);
		}
	}

	// v[u|[g̏ƁA^[Qbgd`FbN
	err = prome_init(clk_mhz);
	if(err)
		escape(err,wait_time);

	/* gK[Zbg[h */
	if(trigger_mode&0x20000000)
	{
		printf("Clear Trigger\n");
		prome_clear_trigger();
		prome_end();
		escape(0,wait_time);
	}

	/* t@C`FbN */
	if( filename[0] == 0)
	{
		printf("\nno file\n");
		prome_end();
		escape(1,wait_time);
	}

	/* ^Cv`FbN */
	if( rom_type < 0)
	{
		printf("\nBad ROM type\n");
		prome_end();
		escape(1,wait_time);
	}

	/* pbhR[hŃobt@𖄂߂Ă */
	memset(data_buf,pad_data,rom_size[rom_type]);

	/* f[^[h */
	file_size = file_read(filename , data_buf , rom_size[rom_type] );

	if(file_size == 0)
	{
		printf("Can't read File %s\n",filename);
		escape(1,wait_time);
	}

	printf("File name '%s' : size %ld bytes\n",filename , file_size);

	/* t@CTCYDWORDE */
	file_size = (file_size+3)&0xfffffffc;

	/* ROM̑S̈ƂɃTCYύX */
	if(all_mode)
		file_size = rom_size[rom_type];

	/* ROM݃[hݒ */
	prome_run_mode(FALSE,rom_type,trigger_mode,reset_on);

	if(ver_mode)
		jmem_verify(0,data_buf,file_size);
	else
	{
		for(retry=RETRY_CNT;retry;retry--)
		{
			printf("\nWriting:");
			jmem_write(0,data_buf,file_size);
			printf("\nVerifying:");
			if(!jmem_verify(0,data_buf,file_size))
				break;
		}
		if(!retry)
			err = 1;
	}
	printf("\n");

	/* gK[AhX\ */
	if(trigger_mode&0x80000000)
	{
		printf("Trigger = %06lxH:%s\n",trigger_mode&0x00ffffff , trigger_mode&0x40000000 ? "PULSE":"SINGLE");
	}

	/* s */
	prome_run_mode(TRUE,rom_type,trigger_mode,reset_off);
	printf("Emulator Start (RESET=%c)\n",reset_off&0xdf);

	prome_end();
	escape(err,wait_time);
}

